@@ -1,6 +1,5 @@
 use adler32::RollingAdler32;
-use crc32fast;
-use std::fmt;
+use core::fmt;
 
 pub struct Adler32(RollingAdler32);
 impl Adler32 {
@@ -39,7 +38,7 @@ impl fmt::Debug for Crc32 {
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     use super::*;
 
     #[test]
