@@ -1,11 +1,10 @@
-use std::cmp;
-use std::io;
-use std::iter;
-use std::ops::Range;
-
-use bit;
-use huffman;
-use huffman::Builder;
+use crate::bit;
+use crate::huffman;
+use crate::huffman::Builder;
+use crate::lz77;
+use alloc::{boxed::Box, vec::Vec};
+use core::{cmp, iter, ops::Range};
+use core2::io;
 
 const FIXED_LITERAL_OR_LENGTH_CODE_TABLE: [(u8, Range<u16>, u16); 4] = [
     (8, 000..144, 0b0_0011_0000),
@@ -90,35 +89,34 @@ const DISTANCE_TABLE: [(u16, u8); 30] = [
 #[derive(Debug, PartialEq, Eq)]
 pub enum Symbol {
     EndOfBlock,
-    Literal(u8),
-    Share { length: u16, distance: u16 },
+    Code(lz77::Code),
 }
 impl Symbol {
     pub fn code(&self) -> u16 {
         match *self {
-            Symbol::Literal(b) => u16::from(b),
+            Symbol::Code(lz77::Code::Literal(b)) => u16::from(b),
             Symbol::EndOfBlock => 256,
-            Symbol::Share { length, .. } => match length {
-                3...10 => 257 + length - 3,
-                11...18 => 265 + (length - 11) / 2,
-                19...34 => 269 + (length - 19) / 4,
-                35...66 => 273 + (length - 35) / 8,
-                67...130 => 277 + (length - 67) / 16,
-                131...257 => 281 + (length - 131) / 32,
+            Symbol::Code(lz77::Code::Pointer { length, .. }) => match length {
+                3..=10 => 257 + length - 3,
+                11..=18 => 265 + (length - 11) / 2,
+                19..=34 => 269 + (length - 19) / 4,
+                35..=66 => 273 + (length - 35) / 8,
+                67..=130 => 277 + (length - 67) / 16,
+                131..=257 => 281 + (length - 131) / 32,
                 258 => 285,
                 _ => unreachable!(),
             },
         }
     }
     pub fn extra_lengh(&self) -> Option<(u8, u16)> {
-        if let Symbol::Share { length, .. } = *self {
+        if let Symbol::Code(lz77::Code::Pointer { length, .. }) = *self {
             match length {
-                3...10 | 258 => None,
-                11...18 => Some((1, (length - 11) % 2)),
-                19...34 => Some((2, (length - 19) % 4)),
-                35...66 => Some((3, (length - 35) % 8)),
-                67...130 => Some((4, (length - 67) % 16)),
-                131...257 => Some((5, (length - 131) % 32)),
+                3..=10 | 258 => None,
+                11..=18 => Some((1, (length - 11) % 2)),
+                19..=34 => Some((2, (length - 19) % 4)),
+                35..=66 => Some((3, (length - 35) % 8)),
+                67..=130 => Some((4, (length - 67) % 16)),
+                131..=257 => Some((5, (length - 131) % 32)),
                 _ => unreachable!(),
             }
         } else {
@@ -126,7 +124,11 @@ impl Symbol {
         }
     }
     pub fn distance(&self) -> Option<(u8, u8, u16)> {
-        if let Symbol::Share { distance, .. } = *self {
+        if let Symbol::Code(lz77::Code::Pointer {
+            backward_distance: distance,
+            ..
+        }) = *self
+        {
             if distance <= 4 {
                 Some((distance as u8 - 1, 0, 0))
             } else {
@@ -151,6 +153,11 @@ impl Symbol {
         }
     }
 }
+impl From<lz77::Code> for Symbol {
+    fn from(code: lz77::Code) -> Self {
+        Symbol::Code(code)
+    }
+}
 
 #[derive(Debug)]
 pub struct Encoder {
@@ -188,11 +195,12 @@ impl Decoder {
         R: io::Read,
     {
         let mut symbol = self.decode_literal_or_length(reader);
-        if let Symbol::Share {
-            ref mut distance, ..
-        } = symbol
+        if let Symbol::Code(lz77::Code::Pointer {
+            ref mut backward_distance,
+            ..
+        }) = symbol
         {
-            *distance = self.decode_distance(reader);
+            *backward_distance = self.decode_distance(reader);
         }
         symbol
     }
@@ -203,20 +211,23 @@ impl Decoder {
     {
         let decoded = self.literal.decode_unchecked(reader);
         match decoded {
-            0...255 => Symbol::Literal(decoded as u8),
+            0..=255 => Symbol::Code(lz77::Code::Literal(decoded as u8)),
             256 => Symbol::EndOfBlock,
             286 | 287 => {
-                let message = format!("The value {} must not occur in compressed data", decoded);
+                #[cfg(feature = "std")]
+                let message = format!("The value {decoded} must not occur in compressed data");
+                #[cfg(not(feature = "std"))]
+                let message = "The value(s) [286, 287] must not occur in compressed data";
                 reader.set_last_error(io::Error::new(io::ErrorKind::InvalidData, message));
                 Symbol::EndOfBlock // dummy value
             }
             length_code => {
                 let (base, extra_bits) = LENGTH_TABLE[length_code as usize - 257];
                 let extra = reader.read_bits_unchecked(extra_bits);
-                Symbol::Share {
+                Symbol::Code(lz77::Code::Pointer {
                     length: base + extra,
-                    distance: 0,
-                }
+                    backward_distance: 0,
+                })
             }
         }
     }
@@ -280,7 +291,7 @@ impl HuffmanCodec for FixedHuffmanCodec {
     where
         R: io::Read,
     {
-        let mut literal_builder = huffman::DecoderBuilder::new(9, Some(END_OF_BLOCK));
+        let mut literal_builder = huffman::DecoderBuilder::new(9, None, Some(END_OF_BLOCK));
         for &(bitwidth, ref symbols, code_base) in &FIXED_LITERAL_OR_LENGTH_CODE_TABLE {
             for (code, symbol) in symbols
                 .clone()
@@ -291,7 +302,8 @@ impl HuffmanCodec for FixedHuffmanCodec {
             }
         }
 
-        let mut distance_builder = huffman::DecoderBuilder::new(5, None);
+        let mut distance_builder =
+            huffman::DecoderBuilder::new(5, literal_builder.safely_peek_bitwidth(), None);
         for i in 0..30 {
             distance_builder.set_mapping(i, huffman::Code::new(5, i))?;
         }
@@ -347,7 +359,7 @@ impl HuffmanCodec for DynamicHuffmanCodec {
             BITWIDTH_CODE_ORDER
                 .iter()
                 .rev()
-                .position(|&i| bitwidth_encoder.lookup(i as u16).width > 0)
+                .position(|&i| code_counts[i] != 0 && bitwidth_encoder.lookup(i as u16).width > 0)
                 .map_or(0, |trailing_zeros| 19 - trailing_zeros),
         ) as u16;
         writer.write_bits(5, literal_code_count - 257)?;
@@ -381,10 +393,12 @@ impl HuffmanCodec for DynamicHuffmanCodec {
         let bitwidth_code_count = reader.read_bits(4)? + 4;
 
         if distance_code_count as usize > MAX_DISTANCE_CODE_COUNT {
+            #[cfg(feature = "std")]
             let message = format!(
-                "The value of HDIST is too big: max={}, actual={}",
-                MAX_DISTANCE_CODE_COUNT, distance_code_count
+                "The value of HDIST is too big: max={MAX_DISTANCE_CODE_COUNT}, actual={distance_code_count}"
             );
+            #[cfg(not(feature = "std"))]
+            let message = "The value of HDIST is too big: max={MAX_DISTANCE_CODE_COUNT}";
             return Err(io::Error::new(io::ErrorKind::InvalidData, message));
         }
 
@@ -396,7 +410,7 @@ impl HuffmanCodec for DynamicHuffmanCodec {
             bitwidth_code_bitwidthes[i] = reader.read_bits(3)? as u8;
         }
         let bitwidth_decoder =
-            huffman::DecoderBuilder::from_bitwidthes(&bitwidth_code_bitwidthes, None)?;
+            huffman::DecoderBuilder::from_bitwidthes(&bitwidth_code_bitwidthes, Some(1), None)?;
 
         let mut literal_code_bitwidthes = Vec::with_capacity(literal_code_count as usize);
         while literal_code_bitwidthes.len() < literal_code_count as usize {
@@ -417,21 +431,28 @@ impl HuffmanCodec for DynamicHuffmanCodec {
             distance_code_bitwidthes.extend(load_bitwidthes(reader, c, last)?);
         }
         if distance_code_bitwidthes.len() > distance_code_count as usize {
+            #[cfg(feature = "std")]
             let message = format!(
                 "The length of `distance_code_bitwidthes` is too large: actual={}, expected={}",
                 distance_code_bitwidthes.len(),
                 distance_code_count
             );
+            #[cfg(not(feature = "std"))]
+            let message = "The length of `distance_code_bitwidthes` is too large";
             return Err(io::Error::new(io::ErrorKind::InvalidData, message));
         }
 
-        Ok(Decoder {
-            literal: huffman::DecoderBuilder::from_bitwidthes(
-                &literal_code_bitwidthes,
-                Some(END_OF_BLOCK),
-            )?,
-            distance: huffman::DecoderBuilder::from_bitwidthes(&distance_code_bitwidthes, None)?,
-        })
+        let literal = huffman::DecoderBuilder::from_bitwidthes(
+            &literal_code_bitwidthes,
+            None,
+            Some(END_OF_BLOCK),
+        )?;
+        let distance = huffman::DecoderBuilder::from_bitwidthes(
+            &distance_code_bitwidthes,
+            Some(literal.safely_peek_bitwidth()),
+            None,
+        )?;
+        Ok(Decoder { literal, distance })
     }
 }
 
@@ -439,12 +460,12 @@ fn load_bitwidthes<R>(
     reader: &mut bit::BitReader<R>,
     code: u16,
     last: Option<u8>,
-) -> io::Result<Box<Iterator<Item = u8>>>
+) -> io::Result<Box<dyn Iterator<Item = u8>>>
 where
     R: io::Read,
 {
     Ok(match code {
-        0...15 => Box::new(iter::once(code as u8)),
+        0..=15 => Box::new(iter::once(code as u8)),
         16 => {
             let count = reader.read_bits(2)? + 3;
             let last = last.ok_or_else(|| invalid_data_error!("No preceding value"))?;
@@ -477,7 +498,7 @@ fn build_bitwidth_codes(
         (&codec.literal, literal_code_count),
         (&codec.distance, distance_code_count),
     ] {
-        for (i, c) in (0..size).map(|x| e.lookup(x as u16).width).enumerate() {
+        for (i, c) in (0..size).map(|x| e.lookup(x).width).enumerate() {
             if i > 0 && run_lens.last().map_or(false, |s| s.value == c) {
                 run_lens.last_mut().unwrap().count += 1;
             } else {
