@@ -4,12 +4,12 @@
 //!
 //! # Examples
 //! ```
-//! use std::io::{self, Read};
+//! use core2::io::{Read, Write};
 //! use libflate::gzip::{Encoder, Decoder};
 //!
 //! // Encoding
 //! let mut encoder = Encoder::new(Vec::new()).unwrap();
-//! io::copy(&mut &b"Hello World!"[..], &mut encoder).unwrap();
+//! encoder.write_all(b"Hello World!".as_ref()).unwrap();
 //! let encoded_data = encoder.finish().into_result().unwrap();
 //!
 //! // Decoding
@@ -19,18 +19,15 @@
 //!
 //! assert_eq!(decoded_data, b"Hello World!");
 //! ```
-use byteorder::LittleEndian;
-use byteorder::ReadBytesExt;
-use byteorder::WriteBytesExt;
-use std::ffi::CString;
-use std::io;
+use crate::checksum;
+use crate::deflate;
+use crate::finish::{Complete, Finish};
+use crate::lz77;
+use alloc::{ffi::CString, vec::Vec};
+use core2::io;
+#[cfg(feature = "std")]
 use std::time;
 
-use checksum;
-use deflate;
-use finish::{Complete, Finish};
-use lz77;
-
 const GZIP_ID: [u8; 2] = [31, 139];
 const COMPRESSION_METHOD_DEFLATE: u8 = 8;
 
@@ -107,17 +104,19 @@ impl Trailer {
     where
         R: io::Read,
     {
-        Ok(Trailer {
-            crc32: reader.read_u32::<LittleEndian>()?,
-            input_size: reader.read_u32::<LittleEndian>()?,
-        })
+        let mut buf = [0; 4];
+        reader.read_exact(&mut buf)?;
+        let crc32 = u32::from_le_bytes(buf);
+        reader.read_exact(&mut buf)?;
+        let input_size = u32::from_le_bytes(buf);
+        Ok(Trailer { crc32, input_size })
     }
     fn write_to<W>(&self, mut writer: W) -> io::Result<()>
     where
         W: io::Write,
     {
-        writer.write_u32::<LittleEndian>(self.crc32)?;
-        writer.write_u32::<LittleEndian>(self.input_size)?;
+        writer.write_all(&self.crc32.to_le_bytes())?;
+        writer.write_all(&self.input_size.to_le_bytes())?;
         Ok(())
     }
 }
@@ -145,12 +144,12 @@ impl HeaderBuilder {
     /// ```
     pub fn new() -> Self {
         // wasm-unknown-unknown does not implement the time module
-        #[cfg(not(target_arch = "wasm32"))]
+        #[cfg(all(not(target_arch = "wasm32"), feature = "std"))]
         let modification_time = time::UNIX_EPOCH
             .elapsed()
             .map(|d| d.as_secs() as u32)
             .unwrap_or(0);
-        #[cfg(target_arch = "wasm32")]
+        #[cfg(any(target_arch = "wasm32", not(feature = "std")))]
         let modification_time = 0;
 
         let header = Header {
@@ -241,6 +240,11 @@ impl HeaderBuilder {
     ///
     /// # Examples
     /// ```
+    /// #[cfg(not(feature = "std"))]
+    /// extern crate alloc;
+    /// #[cfg(not(feature = "std"))]
+    /// use alloc::ffi::CString;
+    /// #[cfg(feature = "std")]
     /// use std::ffi::CString;
     /// use libflate::gzip::HeaderBuilder;
     ///
@@ -256,6 +260,11 @@ impl HeaderBuilder {
     ///
     /// # Examples
     /// ```
+    /// #[cfg(not(feature = "std"))]
+    /// extern crate alloc;
+    /// #[cfg(not(feature = "std"))]
+    /// use alloc::ffi::CString;
+    /// #[cfg(feature = "std")]
     /// use std::ffi::CString;
     /// use libflate::gzip::HeaderBuilder;
     ///
@@ -361,11 +370,9 @@ impl Header {
         W: io::Write,
     {
         writer.write_all(&GZIP_ID)?;
-        writer.write_u8(COMPRESSION_METHOD_DEFLATE)?;
-        writer.write_u8(self.flags())?;
-        writer.write_u32::<LittleEndian>(self.modification_time)?;
-        writer.write_u8(self.compression_level.to_u8())?;
-        writer.write_u8(self.os.to_u8())?;
+        writer.write_all(&[COMPRESSION_METHOD_DEFLATE, self.flags()])?;
+        writer.write_all(&self.modification_time.to_le_bytes())?;
+        writer.write_all(&[self.compression_level.to_u8(), self.os.to_u8()])?;
         if let Some(ref x) = self.extra_field {
             x.write_to(&mut writer)?;
         }
@@ -376,7 +383,7 @@ impl Header {
             writer.write_all(x.as_bytes_with_nul())?;
         }
         if self.is_verified {
-            writer.write_u16::<LittleEndian>(self.crc16())?;
+            writer.write_all(&self.crc16().to_le_bytes())?;
         }
         Ok(())
     }
@@ -385,8 +392,9 @@ impl Header {
         R: io::Read,
     {
         let mut this = HeaderBuilder::new().finish();
-        let mut id = [0; 2];
-        reader.read_exact(&mut id)?;
+        let mut buf = [0; 2 + 1 + 1 + 4 + 1 + 1];
+        reader.read_exact(&mut buf)?;
+        let id = &buf[0..2];
         if id != GZIP_ID {
             return Err(invalid_data_error!(
                 "Unexpected GZIP ID: value={:?}, \
@@ -395,7 +403,7 @@ impl Header {
                 GZIP_ID
             ));
         }
-        let compression_method = reader.read_u8()?;
+        let compression_method = buf[2];
         if compression_method != COMPRESSION_METHOD_DEFLATE {
             return Err(invalid_data_error!(
                 "Compression methods other than DEFLATE(8) are \
@@ -403,10 +411,10 @@ impl Header {
                 compression_method
             ));
         }
-        let flags = reader.read_u8()?;
-        this.modification_time = reader.read_u32::<LittleEndian>()?;
-        this.compression_level = CompressionLevel::from_u8(reader.read_u8()?);
-        this.os = Os::from_u8(reader.read_u8()?);
+        let flags = buf[3];
+        this.modification_time = u32::from_le_bytes([buf[4], buf[5], buf[6], buf[7]]);
+        this.compression_level = CompressionLevel::from_u8(buf[8]);
+        this.os = Os::from_u8(buf[9]);
         if flags & F_EXTRA != 0 {
             this.extra_field = Some(ExtraField::read_from(&mut reader)?);
         }
@@ -420,7 +428,9 @@ impl Header {
         // so that random data from fuzzer can reach actually interesting code.
         // Compilation flag 'fuzzing' is automatically set by all 3 Rust fuzzers.
         if flags & F_HCRC != 0 && cfg!(not(fuzzing)) {
-            let crc = reader.read_u16::<LittleEndian>()?;
+            let mut buf = [0; 2];
+            reader.read_exact(&mut buf)?;
+            let crc = u16::from_le_bytes(buf);
             let expected = this.crc16();
             if crc != expected {
                 return Err(invalid_data_error!(
@@ -442,11 +452,12 @@ where
 {
     let mut buf = Vec::new();
     loop {
-        let b = reader.read_u8()?;
-        if b == 0 {
-            return Ok(unsafe { CString::from_vec_unchecked(buf) });
+        let mut cbuf = [0; 1];
+        reader.read_exact(&mut cbuf)?;
+        if cbuf[0] == 0 {
+            return Ok(CString::new(buf).unwrap());
         }
-        buf.push(b);
+        buf.push(cbuf[0]);
     }
 }
 
@@ -462,7 +473,9 @@ impl ExtraField {
         R: io::Read,
     {
         let mut subfields = Vec::new();
-        let data_size = reader.read_u16::<LittleEndian>()? as usize;
+        let mut buf = [0; 2];
+        reader.read_exact(&mut buf)?;
+        let data_size = u16::from_le_bytes(buf) as usize;
         let mut reader = reader.take(data_size as u64);
         while reader.limit() > 0 {
             subfields.push(ExtraSubField::read_from(&mut reader)?);
@@ -477,7 +490,7 @@ impl ExtraField {
         if len > 0xFFFF {
             return Err(invalid_data_error!("extra field too long: {}", len));
         }
-        writer.write_u16::<LittleEndian>(len as u16)?;
+        writer.write_all(&(len as u16).to_le_bytes())?;
         for f in &self.subfields {
             f.write_to(&mut writer)?;
         }
@@ -505,7 +518,9 @@ impl ExtraSubField {
         };
 
         reader.read_exact(&mut field.id)?;
-        let data_size = reader.read_u16::<LittleEndian>()? as usize;
+        let mut buf = [0; 2];
+        reader.read_exact(&mut buf)?;
+        let data_size = u16::from_le_bytes(buf) as usize;
         field.data.resize(data_size, 0);
         reader.read_exact(&mut field.data)?;
 
@@ -516,7 +531,7 @@ impl ExtraSubField {
         W: io::Write,
     {
         writer.write_all(&self.id)?;
-        writer.write_u16::<LittleEndian>(self.data.len() as u16)?;
+        writer.write_all(&(self.data.len() as u16).to_le_bytes())?;
         writer.write_all(&self.data)?;
         Ok(())
     }
@@ -752,11 +767,11 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Write;
+    /// use core2::io::Write;
     /// use libflate::gzip::Encoder;
     ///
     /// let mut encoder = Encoder::new(Vec::new()).unwrap();
-    /// encoder.write_all(b"Hello World!").unwrap();
+    /// encoder.write_all(&b"Hello World!"[..]).unwrap();
     /// encoder.finish().into_result().unwrap();
     /// ```
     pub fn new(inner: W) -> io::Result<Self> {
@@ -774,13 +789,13 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Write;
+    /// use core2::io::Write;
     /// use libflate::gzip::{Encoder, EncodeOptions, HeaderBuilder};
     ///
     /// let header = HeaderBuilder::new().modification_time(123).finish();
     /// let options = EncodeOptions::new().no_compression().header(header);
     /// let mut encoder = Encoder::with_options(Vec::new(), options).unwrap();
-    /// encoder.write_all(b"Hello World!").unwrap();
+    /// encoder.write_all(&b"Hello World!"[..]).unwrap();
     ///
     /// assert_eq!(encoder.finish().into_result().unwrap(),
     ///            &[31, 139, 8, 0, 123, 0, 0, 0, 0, 3, 1, 12, 0, 243, 255, 72, 101, 108, 108,
@@ -813,11 +828,11 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Write;
+    /// use core2::io::Write;
     /// use libflate::gzip::Encoder;
     ///
     /// let mut encoder = Encoder::new(Vec::new()).unwrap();
-    /// encoder.write_all(b"Hello World!").unwrap();
+    /// encoder.write_all(&b"Hello World!"[..]).unwrap();
     ///
     /// assert!(encoder.finish().as_result().is_ok())
     /// ```
@@ -828,14 +843,17 @@ where
     /// it may be convenient to use `AutoFinishUnchecked` instead of the explicit invocation of this method.
     ///
     /// ```
-    /// use std::io;
+    /// use core2::io::Write;
     /// use libflate::finish::AutoFinishUnchecked;
     /// use libflate::gzip::Encoder;
     ///
     /// let plain = b"Hello World!";
     /// let mut buf = Vec::new();
     /// let mut encoder = AutoFinishUnchecked::new(Encoder::new(&mut buf).unwrap());
-    /// io::copy(&mut &plain[..], &mut encoder).unwrap();
+    /// #[cfg(not(feature = "std"))]
+    /// encoder.write_all(plain.as_ref()).unwrap();
+    /// #[cfg(feature = "std")]
+    /// std::io::copy(&mut &plain[..], &mut encoder).unwrap();
     /// ```
     pub fn finish(self) -> Finish<W, io::Error> {
         let trailer = Trailer {
@@ -907,7 +925,7 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Read;
+    /// use core2::io::Read;
     /// use libflate::gzip::Decoder;
     ///
     /// let encoded_data = [31, 139, 8, 0, 123, 0, 0, 0, 0, 3, 1, 12, 0, 243, 255,
@@ -956,7 +974,7 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Cursor;
+    /// use core2::io::Cursor;
     /// use libflate::gzip::Decoder;
     ///
     /// let encoded_data = [31, 139, 8, 0, 123, 0, 0, 0, 0, 3, 1, 12, 0, 243, 255,
@@ -970,6 +988,13 @@ where
         self.reader.into_inner()
     }
 
+    /// Returns the data that has been decoded but has not yet been read.
+    ///
+    /// This method is useful to retrieve partial decoded data when the decoding process is failed.
+    pub fn unread_decoded_data(&self) -> &[u8] {
+        self.reader.unread_decoded_data()
+    }
+
     fn with_header(inner: R, header: Header) -> Self {
         Decoder {
             header,
@@ -978,6 +1003,13 @@ where
             eos: false,
         }
     }
+
+    fn reset(&mut self, header: Header) {
+        self.header = header;
+        self.reader.reset();
+        self.crc32 = checksum::Crc32::new();
+        self.eos = false;
+    }
 }
 impl<R> io::Read for Decoder<R>
 where
@@ -990,6 +1022,10 @@ where
             let read_size = self.reader.read(buf)?;
             self.crc32.update(&buf[..read_size]);
             if read_size == 0 {
+                if buf.is_empty() {
+                    return Ok(0);
+                }
+
                 self.eos = true;
                 let trailer = Trailer::read_from(self.reader.as_inner_mut())?;
                 // checksum verification is skipped during fuzzing
@@ -1014,8 +1050,8 @@ where
 /// A decoder that decodes all members in a GZIP stream.
 #[derive(Debug)]
 pub struct MultiDecoder<R> {
-    header: Header,
-    decoder: Result<Decoder<R>, R>,
+    decoder: Decoder<R>,
+    eos: bool,
 }
 impl<R> MultiDecoder<R>
 where
@@ -1027,7 +1063,7 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Read;
+    /// use core2::io::Read;
     /// use libflate::gzip::MultiDecoder;
     ///
     /// let mut encoded_data = Vec::new();
@@ -1049,8 +1085,8 @@ where
     pub fn new(inner: R) -> io::Result<Self> {
         let decoder = Decoder::new(inner)?;
         Ok(MultiDecoder {
-            header: decoder.header().clone(),
-            decoder: Ok(decoder),
+            decoder,
+            eos: false,
         })
     }
 
@@ -1068,30 +1104,24 @@ where
     /// assert_eq!(decoder.header().os(), Os::Unix);
     /// ```
     pub fn header(&self) -> &Header {
-        &self.header
+        self.decoder.header()
     }
 
     /// Returns the immutable reference to the inner stream.
     pub fn as_inner_ref(&self) -> &R {
-        match self.decoder {
-            Err(ref reader) => reader,
-            Ok(ref decoder) => decoder.as_inner_ref(),
-        }
+        self.decoder.as_inner_ref()
     }
 
     /// Returns the mutable reference to the inner stream.
     pub fn as_inner_mut(&mut self) -> &mut R {
-        match self.decoder {
-            Err(ref mut reader) => reader,
-            Ok(ref mut decoder) => decoder.as_inner_mut(),
-        }
+        self.decoder.as_inner_mut()
     }
 
     /// Unwraps this `MultiDecoder`, returning the underlying reader.
     ///
     /// # Examples
     /// ```
-    /// use std::io::Cursor;
+    /// use core2::io::Cursor;
     /// use libflate::gzip::MultiDecoder;
     ///
     /// let encoded_data = [31, 139, 8, 0, 123, 0, 0, 0, 0, 3, 1, 12, 0, 243, 255,
@@ -1102,10 +1132,7 @@ where
     /// assert_eq!(decoder.into_inner().into_inner(), &encoded_data[..]);
     /// ```
     pub fn into_inner(self) -> R {
-        match self.decoder {
-            Err(reader) => reader,
-            Ok(decoder) => decoder.into_inner(),
-        }
+        self.decoder.into_inner()
     }
 }
 impl<R> io::Read for MultiDecoder<R>
@@ -1113,73 +1140,56 @@ where
     R: io::Read,
 {
     fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
-        let read_size = match self.decoder {
-            Err(_) => return Ok(0),
-            Ok(ref mut decoder) => decoder.read(buf)?,
-        };
-        // take_mut closure must have the type it borrows as return type,
-        // so we put the function return result to this variable instead.
-        // If function logic is correct, these initial values will never be returned.
-        let mut result: io::Result<usize> = Err(io::Error::new(
-            io::ErrorKind::Other,
-            "If you see this error, please report a bug in libflate",
-        ));
+        if self.eos {
+            return Ok(0);
+        }
+
+        let read_size = self.decoder.read(buf)?;
         if read_size == 0 {
-            take_mut::take(self, |mut owned_self| {
-                let mut reader = owned_self
-                    .decoder
-                    .ok()
-                    .take()
-                    .expect("Never fails")
-                    .into_inner();
-                match Header::read_from(&mut reader) {
-                    Err(e) => {
-                        if e.kind() == io::ErrorKind::UnexpectedEof {
-                            result = Ok(0);
-                        } else {
-                            result = Err(e);
-                        }
-                        owned_self.decoder = Err(reader);
-                        owned_self
-                    }
-                    Ok(header) => {
-                        owned_self.header = header.clone();
-                        owned_self.decoder = Ok(Decoder::with_header(reader, header));
-                        result = owned_self.read(buf);
-                        owned_self
+            match Header::read_from(self.as_inner_mut()) {
+                Err(e) => {
+                    if e.kind() == io::ErrorKind::UnexpectedEof {
+                        self.eos = true;
+                        Ok(0)
+                    } else {
+                        Err(e)
                     }
                 }
-            })
+                Ok(header) => {
+                    self.decoder.reset(header);
+                    self.read(buf)
+                }
+            }
         } else {
-            result = Ok(read_size);
+            Ok(read_size)
         }
-        result
     }
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     use super::*;
-    use finish::AutoFinish;
-    use std::io::{self, Write};
+    use crate::finish::AutoFinish;
+    use alloc::{vec, vec::Vec};
+    use core2::io::{Read, Write};
 
     fn decode(buf: &[u8]) -> io::Result<Vec<u8>> {
         let mut decoder = Decoder::new(buf).unwrap();
         let mut buf = Vec::with_capacity(buf.len());
-        io::copy(&mut decoder, &mut buf)?;
+        decoder.read_to_end(&mut buf)?;
         Ok(buf)
     }
 
     fn decode_multi(buf: &[u8]) -> io::Result<Vec<u8>> {
         let mut decoder = MultiDecoder::new(buf).unwrap();
         let mut buf = Vec::with_capacity(buf.len());
-        io::copy(&mut decoder, &mut buf)?;
+        decoder.read_to_end(&mut buf).unwrap();
         Ok(buf)
     }
 
     fn encode(text: &[u8]) -> io::Result<Vec<u8>> {
         let mut encoder = Encoder::new(Vec::new()).unwrap();
-        io::copy(&mut &text[..], &mut encoder).unwrap();
+        encoder.write_all(text).unwrap();
         encoder.finish().into_result()
     }
 
@@ -1187,7 +1197,7 @@ mod test {
     fn encode_works() {
         let plain = b"Hello World! Hello GZIP!!";
         let mut encoder = Encoder::new(Vec::new()).unwrap();
-        io::copy(&mut &plain[..], &mut encoder).unwrap();
+        encoder.write_all(plain.as_ref()).unwrap();
         let encoded = encoder.finish().into_result().unwrap();
         assert_eq!(decode(&encoded).unwrap(), plain);
     }
@@ -1198,14 +1208,14 @@ mod test {
         let mut buf = Vec::new();
         {
             let mut encoder = AutoFinish::new(Encoder::new(&mut buf).unwrap());
-            io::copy(&mut &plain[..], &mut encoder).unwrap();
+            encoder.write_all(plain.as_ref()).unwrap();
         }
         assert_eq!(decode(&buf).unwrap(), plain);
     }
 
     #[test]
     fn multi_decode_works() {
-        use std::iter;
+        use core::iter;
         let text = b"Hello World!";
         let encoded: Vec<u8> = iter::repeat(encode(text).unwrap())
             .take(2)
@@ -1236,6 +1246,17 @@ mod test {
         assert!(decode(&data[..]).is_err());
     }
 
+    #[test]
+    /// See: https://github.com/sile/libflate/issues/61
+    fn issue_61() {
+        let data = encode(b"Hello World").unwrap();
+        let mut decoder = Decoder::new(&data[..]).unwrap();
+        let mut buf = Vec::new();
+        decoder.read(&mut buf).unwrap();
+        decoder.read_to_end(&mut buf).unwrap();
+        assert_eq!(buf, b"Hello World");
+    }
+
     #[test]
     fn extra_field() {
         let f = ExtraField {
@@ -1252,7 +1273,10 @@ mod test {
     }
 
     #[test]
+    #[cfg(feature = "std")]
     fn encode_with_extra_field() {
+        use std::io;
+
         let mut buf = Vec::new();
         let extra_field = ExtraField {
             subfields: vec![ExtraSubField {
