@@ -1,9 +1,8 @@
-use std::cmp;
-/// Length-limited Huffman Codes
-///
-use std::io;
-
-use bit;
+//! Length-limited Huffman Codes.
+use crate::bit;
+use alloc::{vec, vec::Vec};
+use core::cmp;
+use core2::io;
 
 const MAX_BITWIDTH: u8 = 15;
 
@@ -59,30 +58,45 @@ pub trait Builder: Sized {
 pub struct DecoderBuilder {
     table: Vec<u16>,
     eob_symbol: Option<u16>,
-    eob_bitwidth: u8,
+    safely_peek_bitwidth: Option<u8>,
     max_bitwidth: u8,
 }
 impl DecoderBuilder {
-    pub fn new(max_bitwidth: u8, eob_symbol: Option<u16>) -> Self {
+    pub fn new(
+        max_bitwidth: u8,
+        safely_peek_bitwidth: Option<u8>,
+        eob_symbol: Option<u16>,
+    ) -> Self {
         debug_assert!(max_bitwidth <= MAX_BITWIDTH);
         DecoderBuilder {
             table: vec![u16::from(MAX_BITWIDTH) + 1; 1 << max_bitwidth],
             eob_symbol,
-            eob_bitwidth: max_bitwidth,
+            safely_peek_bitwidth,
             max_bitwidth,
         }
     }
-    pub fn from_bitwidthes(bitwidthes: &[u8], eob_symbol: Option<u16>) -> io::Result<Decoder> {
-        let builder = Self::new(bitwidthes.iter().cloned().max().unwrap_or(0), eob_symbol);
+    pub fn from_bitwidthes(
+        bitwidthes: &[u8],
+        safely_peek_bitwidth: Option<u8>,
+        eob_symbol: Option<u16>,
+    ) -> io::Result<Decoder> {
+        let builder = Self::new(
+            bitwidthes.iter().cloned().max().unwrap_or(0),
+            safely_peek_bitwidth,
+            eob_symbol,
+        );
         builder.restore_canonical_huffman_codes(bitwidthes)
     }
+    pub fn safely_peek_bitwidth(&self) -> Option<u8> {
+        self.safely_peek_bitwidth
+    }
 }
 impl Builder for DecoderBuilder {
     type Instance = Decoder;
     fn set_mapping(&mut self, symbol: u16, code: Code) -> io::Result<()> {
         debug_assert!(code.width <= self.max_bitwidth);
         if Some(symbol) == self.eob_symbol {
-            self.eob_bitwidth = code.width;
+            self.safely_peek_bitwidth = Some(code.width);
         }
 
         // `bitwidth` encoded `to` value
@@ -93,10 +107,13 @@ impl Builder for DecoderBuilder {
         for padding in 0..(1 << (self.max_bitwidth - code.width)) {
             let i = ((padding << code.width) | code_be.bits) as usize;
             if self.table[i] != u16::from(MAX_BITWIDTH) + 1 {
+                #[cfg(feature = "std")]
                 let message = format!(
                     "Bit region conflict: i={}, old_value={}, new_value={}, symbol={}, code={:?}",
                     i, self.table[i], value, symbol, code
                 );
+                #[cfg(not(feature = "std"))]
+                let message = "Bit region conflict";
                 return Err(io::Error::new(io::ErrorKind::InvalidData, message));
             }
             self.table[i] = value;
@@ -106,7 +123,10 @@ impl Builder for DecoderBuilder {
     fn finish(self) -> Self::Instance {
         Decoder {
             table: self.table,
-            eob_bitwidth: self.eob_bitwidth,
+            safely_peek_bitwidth: cmp::min(
+                self.max_bitwidth,
+                self.safely_peek_bitwidth.unwrap_or(1),
+            ),
             max_bitwidth: self.max_bitwidth,
         }
     }
@@ -115,10 +135,14 @@ impl Builder for DecoderBuilder {
 #[derive(Debug)]
 pub struct Decoder {
     table: Vec<u16>,
-    eob_bitwidth: u8,
+    safely_peek_bitwidth: u8,
     max_bitwidth: u8,
 }
 impl Decoder {
+    pub fn safely_peek_bitwidth(&self) -> u8 {
+        self.safely_peek_bitwidth
+    }
+
     #[inline(always)]
     pub fn decode<R>(&self, reader: &mut bit::BitReader<R>) -> io::Result<u16>
     where
@@ -134,18 +158,23 @@ impl Decoder {
     where
         R: io::Read,
     {
-        let code = reader.peek_bits_unchecked(self.eob_bitwidth);
-        let mut value = self.table[code as usize];
-        let mut bitwidth = (value & 0b1_1111) as u8;
-        if bitwidth > self.eob_bitwidth {
-            let code = reader.peek_bits_unchecked(self.max_bitwidth);
+        let mut value;
+        let mut bitwidth;
+        let mut peek_bitwidth = self.safely_peek_bitwidth;
+        loop {
+            let code = reader.peek_bits_unchecked(peek_bitwidth);
             value = self.table[code as usize];
             bitwidth = (value & 0b1_1111) as u8;
+            if bitwidth <= peek_bitwidth {
+                break;
+            }
             if bitwidth > self.max_bitwidth {
                 reader.set_last_error(invalid_data_error!("Invalid huffman coded stream"));
+                break;
             }
+            peek_bitwidth = bitwidth;
         }
-        reader.skip_bits(bitwidth as u8);
+        reader.skip_bits(bitwidth);
         value >> 5
     }
 }
@@ -227,14 +256,14 @@ impl Encoder {
 
 #[allow(dead_code)]
 mod ordinary_huffman_codes {
-    use std::cmp;
-    use std::collections::BinaryHeap;
+    use core::cmp;
+    use dary_heap::BinaryHeap;
 
     pub fn calc_optimal_max_bitwidth(frequencies: &[usize]) -> u8 {
         let mut heap = BinaryHeap::new();
         for &freq in frequencies.iter().filter(|&&f| f > 0) {
             let weight = -(freq as isize);
-            heap.push((weight, 0 as u8));
+            heap.push((weight, 0_u8));
         }
         while heap.len() > 1 {
             let (weight1, width1) = heap.pop().unwrap();
@@ -246,7 +275,8 @@ mod ordinary_huffman_codes {
     }
 }
 mod length_limited_huffman_codes {
-    use std::mem;
+    use alloc::{vec, vec::Vec};
+    use core::mem;
 
     #[derive(Debug, Clone)]
     struct Node {
@@ -334,7 +364,7 @@ mod length_limited_huffman_codes {
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     #[test]
     fn it_works() {}
 }
