@@ -1,25 +1,26 @@
 //! A Rust implementation of DEFLATE algorithm and related formats (ZLIB, GZIP).
+
+#![forbid(unsafe_code)]
 #![warn(missing_docs)]
-extern crate adler32;
-extern crate byteorder;
-extern crate crc32fast;
-extern crate rle_decode_fast;
-extern crate take_mut;
+#![cfg_attr(not(feature = "std"), no_std)]
 
 pub use finish::Finish;
+extern crate alloc;
 
 macro_rules! invalid_data_error {
-    ($fmt:expr) => { invalid_data_error!("{}", $fmt) };
+    ($fmt:expr) => {
+        invalid_data_error!($fmt, "")
+    };
     ($fmt:expr, $($arg:tt)*) => {
-        ::std::io::Error::new(::std::io::ErrorKind::InvalidData, format!($fmt, $($arg)*))
-    }
+        ::core2::io::Error::new(::core2::io::ErrorKind::InvalidData, $fmt)
+    };
 }
 
 macro_rules! finish_try {
     ($e:expr) => {
         match $e.unwrap() {
             (inner, None) => inner,
-            (inner, error) => return ::finish::Finish::new(inner, error),
+            (inner, error) => return crate::finish::Finish::new(inner, error),
         }
     };
 }
