@@ -1,13 +1,8 @@
-use byteorder::LittleEndian;
-use byteorder::ReadBytesExt;
-use rle_decode_fast::rle_decode;
-use std::cmp;
-use std::io;
-use std::io::Read;
-
-use deflate::symbol::{self, HuffmanCodec};
-use lz77;
-use non_blocking::transaction::TransactionalBitReader;
+use crate::deflate::symbol::{self, HuffmanCodec};
+use crate::lz77;
+use crate::non_blocking::transaction::TransactionalBitReader;
+use core::cmp;
+use core2::io::{self, Read};
 /// DEFLATE decoder which supports non-blocking I/O.
 #[derive(Debug)]
 pub struct Decoder<R> {
@@ -23,7 +18,7 @@ impl<R: Read> Decoder<R> {
     ///
     /// # Examples
     /// ```
-    /// use std::io::{Cursor, Read};
+    /// use core2::io::{Cursor, Read};
     /// use libflate::non_blocking::deflate::Decoder;
     ///
     /// let encoded_data = [243, 72, 205, 201, 201, 87, 8, 207, 47, 202, 73, 81, 4, 0];
@@ -56,7 +51,7 @@ impl<R: Read> Decoder<R> {
     ///
     /// # Examples
     /// ```
-    /// use std::io::Cursor;
+    /// use core2::io::Cursor;
     /// use libflate::non_blocking::deflate::Decoder;
     ///
     /// let encoded_data = [243, 72, 205, 201, 201, 87, 8, 207, 47, 202, 73, 81, 4, 0];
@@ -99,8 +94,11 @@ impl<R: Read> Read for Decoder<R> {
                 DecoderState::ReadNonCompressedBlockLen => {
                     let len = self.bit_reader.transaction(|r| {
                         r.reset();
-                        let len = r.as_inner_mut().read_u16::<LittleEndian>()?;
-                        let nlen = r.as_inner_mut().read_u16::<LittleEndian>()?;
+                        let mut buf = [0; 2];
+                        r.as_inner_mut().read_exact(&mut buf)?;
+                        let len = u16::from_le_bytes(buf);
+                        r.as_inner_mut().read_exact(&mut buf)?;
+                        let nlen = u16::from_le_bytes(buf);
                         if !len != nlen {
                             Err(invalid_data_error!(
                                 "LEN={} is not the one's complement of NLEN={}",
@@ -111,7 +109,6 @@ impl<R: Read> Read for Decoder<R> {
                             Ok(len)
                         }
                     })?;
-                    self.block_decoder.buffer.reserve(len as usize);
                     DecoderState::ReadNonCompressedBlock { len }
                 }
                 DecoderState::ReadNonCompressedBlock { len: 0 } => {
@@ -172,21 +169,18 @@ enum DecoderState {
 
 #[derive(Debug)]
 struct BlockDecoder {
-    buffer: Vec<u8>,
-    offset: usize,
+    lz77_decoder: lz77::Lz77Decoder,
     eob: bool,
 }
 impl BlockDecoder {
     pub fn new() -> Self {
         BlockDecoder {
-            buffer: Vec::new(),
-            offset: 0,
+            lz77_decoder: lz77::Lz77Decoder::new(),
             eob: false,
         }
     }
     pub fn enter_new_block(&mut self) {
         self.eob = false;
-        self.truncate_old_buffer();
     }
     pub fn decode<R: Read>(
         &mut self,
@@ -198,18 +192,8 @@ impl BlockDecoder {
         }
         while let Some(s) = self.decode_symbol(bit_reader, symbol_decoder)? {
             match s {
-                symbol::Symbol::Literal(b) => {
-                    self.buffer.push(b);
-                }
-                symbol::Symbol::Share { length, distance } => {
-                    if self.buffer.len() < distance as usize {
-                        return Err(invalid_data_error!(
-                            "Too long backword reference: buffer.len={}, distance={}",
-                            self.buffer.len(),
-                            distance
-                        ));
-                    }
-                    rle_decode(&mut self.buffer, usize::from(distance), usize::from(length));
+                symbol::Symbol::Code(code) => {
+                    self.lz77_decoder.decode(code)?;
                 }
                 symbol::Symbol::EndOfBlock => {
                     self.eob = true;
@@ -219,23 +203,9 @@ impl BlockDecoder {
         }
         Ok(())
     }
-    fn truncate_old_buffer(&mut self) {
-        if self.buffer.len() > lz77::MAX_DISTANCE as usize * 4 {
-            let old_len = self.buffer.len();
-            let new_len = lz77::MAX_DISTANCE as usize;
-            {
-                // isolation to please borrow checker
-                let (dst, src) = self.buffer.split_at_mut(old_len - new_len);
-                dst[..new_len].copy_from_slice(src);
-            }
-            self.buffer.truncate(new_len);
-            self.offset = new_len;
-        }
-    }
 
     fn extend(&mut self, buf: &[u8]) {
-        self.buffer.extend_from_slice(buf);
-        self.offset += buf.len();
+        self.lz77_decoder.extend_from_slice(buf);
     }
 
     fn decode_symbol<R: Read>(
@@ -256,11 +226,8 @@ impl BlockDecoder {
 }
 impl Read for BlockDecoder {
     fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
-        if self.offset < self.buffer.len() {
-            let copy_size = cmp::min(buf.len(), self.buffer.len() - self.offset);
-            buf[..copy_size].copy_from_slice(&self.buffer[self.offset..][..copy_size]);
-            self.offset += copy_size;
-            Ok(copy_size)
+        if !self.lz77_decoder.buffer().is_empty() {
+            self.lz77_decoder.read(buf)
         } else if self.eob {
             Ok(0)
         } else {
@@ -270,16 +237,17 @@ impl Read for BlockDecoder {
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     use super::*;
-    use deflate::{EncodeOptions, Encoder};
-    use std::io::{self, Read};
-    use util::{nb_read_to_end, WouldBlockReader};
+    use crate::deflate::{EncodeOptions, Encoder};
+    use crate::util::{nb_read_to_end, WouldBlockReader};
+    use alloc::{format, string::String, vec::Vec};
+    use core2::io::{Read, Write};
 
     #[test]
     fn it_works() {
         let mut encoder = Encoder::new(Vec::new());
-        io::copy(&mut &b"Hello World!"[..], &mut encoder).unwrap();
+        encoder.write_all(b"Hello World!".as_ref()).unwrap();
         let encoded_data = encoder.finish().into_result().unwrap();
 
         let mut decoder = Decoder::new(&encoded_data[..]);
@@ -292,7 +260,7 @@ mod test {
     #[test]
     fn non_blocking_io_works() {
         let mut encoder = Encoder::new(Vec::new());
-        io::copy(&mut &b"Hello World!"[..], &mut encoder).unwrap();
+        encoder.write_all(b"Hello World!".as_ref()).unwrap();
         let encoded_data = encoder.finish().into_result().unwrap();
 
         let decoder = Decoder::new(WouldBlockReader::new(&encoded_data[..]));
@@ -308,8 +276,8 @@ mod test {
             .map(|i| format!("test {}", i))
             .collect();
 
-        let mut encoder = ::deflate::Encoder::new(Vec::new());
-        io::copy(&mut text.as_bytes(), &mut encoder).unwrap();
+        let mut encoder = crate::deflate::Encoder::new(Vec::new());
+        encoder.write_all(text.as_bytes()).unwrap();
         let encoded_data = encoder.finish().into_result().unwrap();
 
         let decoder = Decoder::new(WouldBlockReader::new(&encoded_data[..]));
@@ -320,7 +288,7 @@ mod test {
     #[test]
     fn non_compressed_non_blocking_io_works() {
         let mut encoder = Encoder::with_options(Vec::new(), EncodeOptions::new().no_compression());
-        io::copy(&mut &b"Hello World!"[..], &mut encoder).unwrap();
+        encoder.write_all(b"Hello World!".as_ref()).unwrap();
         let encoded_data = encoder.finish().into_result().unwrap();
 
         let decoder = Decoder::new(WouldBlockReader::new(&encoded_data[..]));
