@@ -4,12 +4,12 @@
 //!
 //! # Examples
 //! ```
-//! use std::io::{self, Read};
+//! use core2::io::{Read, Write};
 //! use libflate::zlib::{Encoder, Decoder};
 //!
 //! // Encoding
 //! let mut encoder = Encoder::new(Vec::new()).unwrap();
-//! io::copy(&mut &b"Hello World!"[..], &mut encoder).unwrap();
+//! encoder.write_all(&b"Hello World!"[..]).unwrap();
 //! let encoded_data = encoder.finish().into_result().unwrap();
 //!
 //! // Decoding
@@ -19,15 +19,11 @@
 //!
 //! assert_eq!(decoded_data, b"Hello World!");
 //! ```
-use byteorder::BigEndian;
-use byteorder::ReadBytesExt;
-use byteorder::WriteBytesExt;
-use std::io;
-
-use checksum;
-use deflate;
-use finish::{Complete, Finish};
-use lz77;
+use crate::checksum;
+use crate::deflate;
+use crate::finish::{Complete, Finish};
+use crate::lz77;
+use core2::io;
 
 const COMPRESSION_METHOD_DEFLATE: u8 = 8;
 
@@ -73,6 +69,8 @@ impl From<lz77::CompressionLevel> for CompressionLevel {
 
 /// LZ77 Window sizes defined by the ZLIB format.
 #[derive(Debug, Clone, PartialOrd, Ord, PartialEq, Eq, Hash)]
+// TODO: Use `#[allow(clippy::upper_case_acronyms)]` instead once it gets available on the stable branch
+#[allow(clippy::all)]
 pub enum Lz77WindowSize {
     /// 256 bytes
     B256 = 0,
@@ -175,6 +173,27 @@ impl Lz77WindowSize {
     }
 }
 
+/// [zlib] library specific parameter for defining behavior when `Write::flush` method is called.
+///
+/// # References
+///
+/// - [Zlib Manual](https://www.zlib.net/manual.html)
+/// - [Zlib Flush Modes](https://www.bolet.org/~pornin/deflate-flush.html)
+///
+/// [zlib]: https://www.zlib.net/
+#[derive(Debug, Default, Clone, Copy, PartialEq, Eq, Hash, PartialOrd, Ord)]
+pub enum FlushMode {
+    /// `Z_NO_FLUSH` (default).
+    ///
+    /// Note that when this parameter is specified,
+    /// no `zlib` specific processing will not be executed but ordinal DEFLATE layer flushing will be performed.
+    #[default]
+    None = 0,
+
+    /// `Z_SYNC_FLUSH`.
+    Sync = 2,
+}
+
 /// ZLIB header.
 #[derive(Debug, Clone, PartialEq, Eq, Hash)]
 pub struct Header {
@@ -203,8 +222,9 @@ impl Header {
     where
         R: io::Read,
     {
-        let cmf = reader.read_u8()?;
-        let flg = reader.read_u8()?;
+        let mut buf = [0; 2];
+        reader.read_exact(&mut buf)?;
+        let [cmf, flg] = buf;
         let check = (u16::from(cmf) << 8) + u16::from(flg);
         if check % 31 != 0 {
             return Err(invalid_data_error!(
@@ -228,13 +248,14 @@ impl Header {
             invalid_data_error!("CINFO above 7 are not allowed: value={}", compression_info)
         })?;
 
-        let dict_flag = (flg & 0b100_000) != 0;
+        let dict_flag = (flg & 0b10_0000) != 0;
         if dict_flag {
-            let dictionary_id = reader.read_u32::<BigEndian>()?;
+            let mut buf = [0; 4];
+            reader.read_exact(&mut buf)?;
             return Err(invalid_data_error!(
                 "Preset dictionaries are not supported: \
                  dictionary_id=0x{:X}",
-                dictionary_id
+                u32::from_be_bytes(buf)
             ));
         }
         let compression_level = CompressionLevel::from_u2(flg >> 6);
@@ -253,8 +274,7 @@ impl Header {
         if check % 31 != 0 {
             flg += (31 - check % 31) as u8;
         }
-        writer.write_u8(cmf)?;
-        writer.write_u8(flg)?;
+        writer.write_all(&[cmf, flg])?;
         Ok(())
     }
 }
@@ -277,7 +297,7 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Read;
+    /// use core2::io::Read;
     /// use libflate::zlib::Decoder;
     ///
     /// let encoded_data = [120, 156, 243, 72, 205, 201, 201, 87, 8, 207, 47,
@@ -330,7 +350,7 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Cursor;
+    /// use core2::io::Cursor;
     /// use libflate::zlib::Decoder;
     ///
     /// let encoded_data = [120, 156, 243, 72, 205, 201, 201, 87, 8, 207, 47,
@@ -342,6 +362,13 @@ where
     pub fn into_inner(self) -> R {
         self.reader.into_inner()
     }
+
+    /// Returns the data that has been decoded but has not yet been read.
+    ///
+    /// This method is useful to retrieve partial decoded data when the decoding process is failed.
+    pub fn unread_decoded_data(&self) -> &[u8] {
+        self.reader.unread_decoded_data()
+    }
 }
 impl<R> io::Read for Decoder<R>
 where
@@ -353,8 +380,15 @@ where
         } else {
             let read_size = self.reader.read(buf)?;
             if read_size == 0 {
+                if buf.is_empty() {
+                    return Ok(0);
+                }
+
                 self.eos = true;
-                let adler32 = self.reader.as_inner_mut().read_u32::<BigEndian>()?;
+                let mut buf = [0; 4];
+                self.reader.as_inner_mut().read_exact(&mut buf)?;
+                let adler32 = u32::from_be_bytes(buf);
+
                 // checksum verification is skipped during fuzzing
                 // so that random data from fuzzer can reach actually interesting code
                 // Compilation flag 'fuzzing' is automatically set by all 3 Rust fuzzers.
@@ -383,12 +417,14 @@ where
 {
     header: Header,
     options: deflate::EncodeOptions<E>,
+    flush_mode: FlushMode,
 }
 impl Default for EncodeOptions<lz77::DefaultLz77Encoder> {
     fn default() -> Self {
         EncodeOptions {
             header: Header::from_lz77(&lz77::DefaultLz77Encoder::new()),
             options: Default::default(),
+            flush_mode: FlushMode::None,
         }
     }
 }
@@ -424,6 +460,7 @@ where
         EncodeOptions {
             header: Header::from_lz77(&lz77),
             options: deflate::EncodeOptions::with_lz77(lz77),
+            flush_mode: FlushMode::None,
         }
     }
 
@@ -472,12 +509,19 @@ where
         self.options = self.options.fixed_huffman_codes();
         self
     }
+
+    /// Specifies flush mode.
+    pub fn flush_mode(mut self, mode: FlushMode) -> Self {
+        self.flush_mode = mode;
+        self
+    }
 }
 
 /// ZLIB encoder.
 #[derive(Debug)]
 pub struct Encoder<W, E = lz77::DefaultLz77Encoder> {
     header: Header,
+    flush_mode: FlushMode,
     writer: deflate::Encoder<W, E>,
     adler32: checksum::Adler32,
 }
@@ -491,7 +535,10 @@ where
     ///
     /// # Examples
     /// ```
+    /// #[cfg(feature = "std")]
     /// use std::io::Write;
+    /// #[cfg(not(feature = "std"))]
+    /// use core2::io::Write;
     /// use libflate::zlib::Encoder;
     ///
     /// let mut encoder = Encoder::new(Vec::new()).unwrap();
@@ -516,7 +563,7 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Write;
+    /// use core2::io::Write;
     /// use libflate::zlib::{Encoder, EncodeOptions};
     ///
     /// let options = EncodeOptions::new().no_compression();
@@ -531,6 +578,7 @@ where
         options.header.write_to(&mut inner)?;
         Ok(Encoder {
             header: options.header,
+            flush_mode: options.flush_mode,
             writer: deflate::Encoder::with_options(inner, options.options),
             adler32: checksum::Adler32::new(),
         })
@@ -553,7 +601,7 @@ where
     ///
     /// # Examples
     /// ```
-    /// use std::io::Write;
+    /// use core2::io::Write;
     /// use libflate::zlib::Encoder;
     ///
     /// let mut encoder = Encoder::new(Vec::new()).unwrap();
@@ -570,19 +618,19 @@ where
     /// it may be convenient to use `AutoFinishUnchecked` instead of the explicit invocation of this method.
     ///
     /// ```
-    /// use std::io;
+    /// use core2::io::Write;
     /// use libflate::finish::AutoFinishUnchecked;
     /// use libflate::zlib::Encoder;
     ///
     /// let plain = b"Hello World!";
     /// let mut buf = Vec::new();
     /// let mut encoder = AutoFinishUnchecked::new(Encoder::new(&mut buf).unwrap());
-    /// io::copy(&mut &plain[..], &mut encoder).unwrap();
+    /// encoder.write_all(plain.as_ref()).unwrap();
     /// ```
     pub fn finish(self) -> Finish<W, io::Error> {
         let mut inner = finish_try!(self.writer.finish());
         match inner
-            .write_u32::<BigEndian>(self.adler32.value())
+            .write_all(&self.adler32.value().to_be_bytes())
             .and_then(|_| inner.flush())
         {
             Ok(_) => Finish::new(inner, None),
@@ -616,7 +664,10 @@ where
         Ok(written_size)
     }
     fn flush(&mut self) -> io::Result<()> {
-        self.writer.flush()
+        match self.flush_mode {
+            FlushMode::None => self.writer.flush(),
+            FlushMode::Sync => self.writer.zlib_sync_flush(),
+        }
     }
 }
 impl<W, E> Complete for Encoder<W, E>
@@ -630,20 +681,21 @@ where
 }
 
 #[cfg(test)]
-mod test {
+mod tests {
     use super::*;
-    use finish::AutoFinish;
-    use std::io;
+    use crate::finish::AutoFinish;
+    use alloc::{borrow::ToOwned, string::ToString, vec, vec::Vec};
+    use core2::io::{Read as _, Write as _};
 
     fn decode_all(buf: &[u8]) -> io::Result<Vec<u8>> {
         let mut decoder = Decoder::new(buf).unwrap();
         let mut buf = Vec::with_capacity(buf.len());
-        io::copy(&mut decoder, &mut buf)?;
+        decoder.read_to_end(&mut buf)?;
         Ok(buf)
     }
     fn default_encode(buf: &[u8]) -> io::Result<Vec<u8>> {
         let mut encoder = Encoder::new(Vec::new()).unwrap();
-        io::copy(&mut &buf[..], &mut encoder).unwrap();
+        encoder.write_all(buf).unwrap();
         encoder.finish().into_result()
     }
     macro_rules! assert_encode_decode {
@@ -669,7 +721,7 @@ mod test {
         );
 
         let mut buf = Vec::new();
-        io::copy(&mut decoder, &mut buf).unwrap();
+        decoder.read_to_end(&mut buf).unwrap();
 
         let expected = b"Hello World!";
         assert_eq!(buf, expected);
@@ -679,7 +731,7 @@ mod test {
     fn default_encode_works() {
         let plain = b"Hello World! Hello ZLIB!!";
         let mut encoder = Encoder::new(Vec::new()).unwrap();
-        io::copy(&mut &plain[..], &mut encoder).unwrap();
+        encoder.write_all(plain.as_ref()).unwrap();
         let encoded = encoder.finish().into_result().unwrap();
         assert_eq!(decode_all(&encoded).unwrap(), plain);
     }
@@ -690,7 +742,7 @@ mod test {
         let mut encoder =
             Encoder::with_options(Vec::new(), EncodeOptions::default().fixed_huffman_codes())
                 .unwrap();
-        io::copy(&mut &plain[..], &mut encoder).unwrap();
+        encoder.write_all(plain.as_ref()).unwrap();
         let encoded = encoder.finish().into_result().unwrap();
         assert_eq!(decode_all(&encoded).unwrap(), plain);
     }
@@ -704,7 +756,7 @@ mod test {
         let plain = b"Hello World!";
         let mut encoder =
             Encoder::with_options(Vec::new(), EncodeOptions::new().no_compression()).unwrap();
-        io::copy(&mut &plain[..], &mut encoder).unwrap();
+        encoder.write_all(plain.as_ref()).unwrap();
         let encoded = encoder.finish().into_result().unwrap();
         let expected = RAW_ENCODE_WORKS_EXPECTED;
         assert_eq!(encoded, expected);
@@ -717,7 +769,7 @@ mod test {
         let mut buf = Vec::new();
         {
             let mut encoder = AutoFinish::new(Encoder::new(&mut buf).unwrap());
-            io::copy(&mut &plain[..], &mut encoder).unwrap();
+            encoder.write_all(plain.as_ref()).unwrap();
         }
         assert_eq!(decode_all(&buf).unwrap(), plain);
     }
@@ -750,22 +802,134 @@ mod test {
         let encoded =
             include_bytes!("../data/issues_16/crash-1bb6d408475a5bd57247ee40f290830adfe2086e");
         assert_eq!(
-            decode_all(&encoded[..]).err().map(|e| e.to_string()),
-            Some("The value of HDIST is too big: max=30, actual=32".to_owned())
+            &decode_all(&encoded[..])
+                .err()
+                .map(|e| e.to_string())
+                .unwrap()[..31],
+            "The value of HDIST is too big: max=30, actual=32"[..31]
+                .to_owned()
+                .as_str()
         );
 
         let encoded =
             include_bytes!("../data/issues_16/crash-369e8509a0e76356f4549c292ceedee429cfe125");
         assert_eq!(
-            decode_all(&encoded[..]).err().map(|e| e.to_string()),
-            Some("The value of HDIST is too big: max=30, actual=32".to_owned())
+            &decode_all(&encoded[..])
+                .err()
+                .map(|e| e.to_string())
+                .unwrap()[..31],
+            "The value of HDIST is too big: max=30, actual=32"[..31]
+                .to_owned()
+                .as_str()
         );
 
         let encoded =
             include_bytes!("../data/issues_16/crash-e75959d935650306881140df7f6d1d73e33425cb");
         assert_eq!(
-            decode_all(&encoded[..]).err().map(|e| e.to_string()),
-            Some("The value of HDIST is too big: max=30, actual=32".to_owned())
+            &decode_all(&encoded[..])
+                .err()
+                .map(|e| e.to_string())
+                .unwrap()[..31],
+            "The value of HDIST is too big: max=30, actual=32"[..31]
+                .to_owned()
+                .as_str()
         );
     }
+
+    #[test]
+    fn test_issues_27() {
+        // See: https://github.com/sile/libflate/issues/27
+
+        let writes = ["fooooooooooooooooo", "bar", "baz"];
+
+        // FlushMode::None
+        let mut encoder = Encoder::new(Vec::new()).unwrap();
+        for _ in 0..2 {
+            for string in &writes {
+                encoder.write(string.as_bytes()).expect("Write failed");
+            }
+            encoder.flush().expect("Flush failed");
+        }
+        let finished = encoder.finish().unwrap();
+        let expected = vec![
+            120, 156, // header
+            92, 192, 161, 17, 0, 0, 0, 1, 192, 89, 9, 170, 59, 209, 244, 186, 151, 31, 17, 162,
+            227, 2, 14, 141, 0, 0, 0, 8, 0, 206, 74, 80, 221, 137, 166, 215, 189, 252, 136, 16, 93,
+            1, 112, 32, 0, 0, 0, 0, 0, 228, 255, 26, 246, 95, 20, 111,
+        ];
+        assert_eq!(finished.0, expected);
+
+        let mut output = Vec::new();
+        Decoder::new(&finished.0[..])
+            .unwrap()
+            .read_to_end(&mut output)
+            .unwrap();
+        assert_eq!(
+            output,
+            "fooooooooooooooooobarbazfooooooooooooooooobarbaz".as_bytes()
+        );
+
+        // FlushMode::Sync
+        let mut encoder =
+            Encoder::with_options(Vec::new(), EncodeOptions::new().flush_mode(FlushMode::Sync))
+                .unwrap();
+        for _ in 0..2 {
+            for string in &writes {
+                encoder.write(string.as_bytes()).expect("Write failed");
+            }
+            encoder.flush().expect("Flush failed");
+        }
+        let finished = encoder.finish().unwrap();
+        let expected = vec![
+            120, 156, // header
+            92, 192, 161, 17, 0, 0, 0, 1, 192, 89, 9, 170, 59, 209, 244, 186, 151, 31, 17, 162, 3,
+            0, 0, 255, 255, // sync bytes
+            92, 192, 161, 17, 0, 0, 0, 1, 192, 89, 9, 170, 59, 209, 244, 186, 151, 31, 17, 162, 3,
+            0, 0, 255, 255, // sync bytes
+            5, 192, 129, 0, 0, 0, 0, 0, 144, 255, 107, 0, 246, 95, 20, 111,
+        ];
+        assert_eq!(finished.0, expected);
+
+        let mut output = Vec::new();
+        Decoder::new(&finished.0[..])
+            .unwrap()
+            .read_to_end(&mut output)
+            .unwrap();
+        assert_eq!(
+            output,
+            "fooooooooooooooooobarbazfooooooooooooooooobarbaz".as_bytes()
+        );
+    }
+
+    #[test]
+    #[cfg(feature = "std")]
+    /// See: https://github.com/sile/libflate/issues/61
+    fn issue_61() {
+        let data = default_encode(b"Hello World").unwrap();
+        let mut decoder = Decoder::new(&data[..]).unwrap();
+        let mut buf = Vec::new();
+        decoder.read(&mut buf).unwrap();
+        decoder.read_to_end(&mut buf).unwrap();
+        assert_eq!(buf, b"Hello World");
+    }
+
+    #[test]
+    fn issue71() {
+        let encoded_data = [
+            120, 218, 251, 255, 207, 144, 193, 138, 193, 151, 161, 146, 33, 143, 33, 149, 161, 156,
+            161, 24, 72, 38, 51, 148, 48, 100, 50, 228, 3, 69, 120, 25, 184, 24,
+        ];
+
+        let mut decoder = Decoder::new(&encoded_data[..]).unwrap();
+        let mut buf = Vec::new();
+        let result = decoder.read_to_end(&mut buf);
+        assert!(result.is_err());
+        buf.extend_from_slice(decoder.unread_decoded_data());
+
+        let decoded_data = [
+            255, 254, 49, 0, 58, 0, 77, 0, 121, 0, 110, 0, 101, 0, 119, 0, 115, 0, 101, 0, 99, 0,
+            116, 0, 105, 0, 111, 0, 110, 0, 13, 0, 10,
+        ];
+        assert_eq!(buf, decoded_data);
+    }
 }
